﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections.Specialized;
using System.Net;
using System.IO;
using System.Text.RegularExpressions;
using System.Drawing;

namespace BSF_Utilities
{
    /// <summary>
    /// The web miner is for finding substrings in a webpage. The methods used to do this are
    /// basic line by line iterated regex lookups. Also relative string look arounds. This is basically
    /// looking for an initial string, then looking around that string (with a specified radius) for 
    /// the actual string to yield.
    /// </summary>
    class WebMiner
    {
        /// <summary>
        /// Parameters to define a look around operation for the web miner
        /// </summary>
        public class LookAroundParams
        {
            public enum LookDirection { Down, Up, Both };
            private string _lookAroundPattern = string.Empty;
            private string _lookForPattern = string.Empty;
            private int _lookAroundRadius = 6;
            private LookDirection _lookDirection;

            public LookAroundParams(string lookAroundPattern, string lookForPattern,
                int lookAroundRadius, LookDirection lookDirection)
            {
                _lookAroundPattern = lookAroundPattern;
                _lookForPattern = lookForPattern;
                _lookAroundRadius = lookAroundRadius;
                _lookDirection = lookDirection;
            }

            /// <summary>
            /// Gets the pattern that is to be found then looked around (for the LookForPattern)
            /// </summary>
            public string LookAroundPattern
            {
                get { return _lookAroundPattern; }
            }

            /// <summary>
            /// Gets the pattern that represents the actual value to be retrieved
            /// </summary>
            public string LookForPattern
            {
                get { return _lookForPattern; }
                set { _lookForPattern = value; }
            }

            /// <summary>
            /// Gets the number of lines to look around the LookAround find
            /// </summary>
            public int LookAroundRadius
            {
                get { return _lookAroundRadius; }
                set { _lookAroundRadius = value; }
            }

            /// <summary>
            /// Gets the direction to search around the LookAround find. Both will
            /// search up AND down 
            /// </summary>
            public LookDirection DirectionTolook
            {
                get { return _lookDirection; }
            }
        }

        public event OpenReadCompletedEventHandler PageDownloadCompleted;

        private string _originalPageLink = string.Empty;
        private string[] _pageToMine = null;
        private LookAroundParams _lookAroundParams;

        public static readonly RegexOptions RecommendedOptions = 
            RegexOptions.IgnoreCase | RegexOptions.CultureInvariant | RegexOptions.Compiled;

        public WebMiner(string pageToMineLink)
        {
            _originalPageLink = pageToMineLink;
        }

        public virtual void DownloadWebpage()
        {
            WebClient webClient = new WebClient();
            webClient.OpenReadCompleted += new OpenReadCompletedEventHandler(webClient_OpenReadCompleted);
            webClient.OpenReadAsync(new Uri(_originalPageLink));
        }

        protected virtual void OnPageDownloadCompleted(OpenReadCompletedEventArgs e)
        {
            if (PageDownloadCompleted != null)
                PageDownloadCompleted(this, e);
        }

        protected virtual void webClient_OpenReadCompleted(object sender, OpenReadCompletedEventArgs e)
		{
			if (e.Error == null)
			{
                StreamReader SR = new StreamReader(e.Result);
                List<string> content = new List<string>();
                string line;

                while ((line = SR.ReadLine()) != null)
                {
                    content.Add(line);
                }

                SR.Dispose();
                _pageToMine = content.ToArray();
                ((WebClient)sender).Dispose();
			}

			OnPageDownloadCompleted(e);
		}

        #region Look For
        /// <summary>
        /// Does a basic iterative lookup for a string. If groups are captured, yield the first group value
        /// </summary>
        public Match LookFor(string pattern, RegexOptions options, out int line)
        {
            Regex regex = new Regex(pattern, options);

            for (int i = 0; i < _pageToMine.Length; i++)
            {
                //if (i >= _pageToMine.Length) return string.Empty;
                
                Match match = regex.Match(_pageToMine[i]);

                if (match.Success)
                {
                    line = i;
                    return match;
                }
            }

            line = -1;
            return null;
        }
#endregion

        #region LookAroundFor
        /// <summary>
        /// Looks around a specified pattern for the actual pattern to return.
        /// If groups are captured, yield the first group value
        /// </summary>
        public string LookAroundFor()
        {
            RegexOptions options = RegexOptions.IgnoreCase | RegexOptions.CultureInvariant | RegexOptions.Compiled;
            Regex regexLookAround = new Regex(_lookAroundParams.LookAroundPattern, options);
            Regex regexLookFor = new Regex(_lookAroundParams.LookForPattern, options);

                for (int i = 0; i < _pageToMine.Length; i++)
                {
                    if (regexLookAround.IsMatch(_pageToMine[i]))
                    {
                        switch (_lookAroundParams.DirectionTolook)
                        {
                            #region Iterative Logic
                            case LookAroundParams.LookDirection.Down:
                                {
                                    for (int i2 = 0; i2 <= _lookAroundParams.LookAroundRadius; i2++)
                                    {
                                        if (i + i2 >= _pageToMine.Length) break;

                                        Match match = regexLookFor.Match(_pageToMine[i + i2]);

                                        if (match.Success)
                                        {
                                            // There seems to always be 1 captured group
                                            // being the pattern itself so check if there is more
                                            // if there is then return the 1st defined groupGobbl
                                            if (match.Groups.Count > 1)
                                            {
                                                return match.Groups[1].Value;
                                            }

                                            return match.Value;
                                        }
                                    }

                                    break;
                                }
                            case LookAroundParams.LookDirection.Up:
                                {
                                    for (int i2 = 0; i2 <= _lookAroundParams.LookAroundRadius; i2++)
                                    {
                                        if (i - i2 < 0) break; //Break if outside index bounds

                                        Match match = regexLookFor.Match(_pageToMine[i - i2]);

                                        if (match.Success)
                                        {
                                            if (match.Groups.Count > 1)
                                            {
                                                return match.Groups[1].Value;
                                            }

                                            return match.Value;
                                        }
                                    }

                                    break;
                                }

                            case LookAroundParams.LookDirection.Both:
                                {
                                    // Look Down
                                    for (int i2 = 0; i2 <= _lookAroundParams.LookAroundRadius; i2++)
                                    {
                                        if (i + i2 >= _pageToMine.Length) break;

                                        Match match = regexLookFor.Match(_pageToMine[i + i2]);

                                        if (match.Success)
                                        {
                                            if (match.Groups.Count > 1)
                                            {
                                                return match.Groups[1].Value;
                                            }

                                            return match.Value;
                                        }
                                    }

                                    // look Up but this time start at i for efficiency, no need to
                                    // scan the bottom half again
                                    for (int i2 = i; i2 >= i - _lookAroundParams.LookAroundRadius; i2--)
                                    {
                                        if (i2 == 0)
                                            break; //Break if outside index bounds

                                        Match match = regexLookFor.Match(_pageToMine[i2]);

                                        if (match.Success)
                                        {
                                            if (match.Groups.Count > 1)
                                            {
                                                return match.Groups[1].Value;
                                            }

                                            return match.Value;
                                        }
                                    }

                                    break;
                                }
                            #endregion
                        }
                    }
                }

            return string.Empty;
        }

        public event LookAroundInteratedEventHandler LookAroundInterated;

        public delegate void LookAroundInteratedEventHandler(object sender, LookAroundInteratedEventArgs e);

        public class LookAroundInteratedEventArgs : EventArgs
        {
            private int _currentLine = 0;

            public LookAroundInteratedEventArgs(int currentLine)
            {
                _currentLine = currentLine;
            }

            public int CurrentLine
            {
                get { return _currentLine; }
            }
        }

        protected virtual void OnLookAroundIterated(LookAroundInteratedEventArgs e)
        {
            if (LookAroundInterated != null)
            {
                LookAroundInterated(this, e);
            }
        }

        /// <summary>
        /// Looks around a specified pattern for the actual pattern to return.
        /// If groups are captured, yield the first group value
        /// </summary>
        public Match LookAroundFor(RegexOptions options)
        {
            Regex regexLookAround = new Regex(_lookAroundParams.LookAroundPattern, options);
            Regex regexLookFor = new Regex(_lookAroundParams.LookForPattern, options);

            for (int i = 0; i < _pageToMine.Length; i++)
            {
                LookAroundInteratedEventArgs eArgs = new LookAroundInteratedEventArgs(i);
                OnLookAroundIterated(eArgs);

                if (regexLookAround.IsMatch(_pageToMine[i]))
                {
                    switch (_lookAroundParams.DirectionTolook)
                    {
                        #region Iterative Logic
                        case LookAroundParams.LookDirection.Down:
                            {
                                for (int i2 = 0; i2 <= _lookAroundParams.LookAroundRadius; i2++)
                                {
                                    if (i + i2 >= _pageToMine.Length) break;

                                    Match match = regexLookFor.Match(_pageToMine[i + i2]);

                                    LookAroundInteratedEventArgs eArgs2 = new LookAroundInteratedEventArgs(i);
                                    OnLookAroundIterated(eArgs2);

                                    if (match.Success)
                                    {
                                        return match;
                                    }
                                }

                                break;
                            }
                        case LookAroundParams.LookDirection.Up:
                            {
                                for (int i2 = 0; i2 <= _lookAroundParams.LookAroundRadius; i2++)
                                {
                                    if (i - i2 < 0) break; //Break if outside index bounds

                                    Match match = regexLookFor.Match(_pageToMine[i - i2]);

                                    LookAroundInteratedEventArgs eArgs2 = new LookAroundInteratedEventArgs(i - i2);
                                    OnLookAroundIterated(eArgs2);

                                    if (match.Success)
                                    {
                                        return match;
                                    }
                                }

                                break;
                            }

                        case LookAroundParams.LookDirection.Both:
                            {
                                // Look Down
                                for (int i2 = 0; i2 <= _lookAroundParams.LookAroundRadius; i2++)
                                {
                                    if (i + i2 >= _pageToMine.Length) break;

                                    Match match = regexLookFor.Match(_pageToMine[i + i2]);

                                    LookAroundInteratedEventArgs eArgs2 = new LookAroundInteratedEventArgs(i + i2);
                                    OnLookAroundIterated(eArgs2);

                                    if (match.Success)
                                    {
                                        return match;
                                    }
                                }

                                // look Up but this time start at i for efficiency, no need to
                                // scan the bottom half again
                                for (int i2 = i; i2 >= i - _lookAroundParams.LookAroundRadius; i2--)
                                {
                                    if (i2 == 0)
                                        break; //Break if outside index bounds

                                    Match match = regexLookFor.Match(_pageToMine[i2]);

                                    LookAroundInteratedEventArgs eArgs2 = new LookAroundInteratedEventArgs(i2);
                                    OnLookAroundIterated(eArgs2);

                                    if (match.Success)
                                    {
                                        return match;
                                    }
                                }



                                break;
                            }

                        #endregion
                    }
                }
            }

            return null;
        }

        /// <summary>
        /// Looks around a specified pattern for the actual pattern to return.
        /// If groups are captured, yield the first group value
        /// </summary>
        public Match LookAroundFor(RegexOptions options, out int line)
        {
            Regex regexLookAround = new Regex(_lookAroundParams.LookAroundPattern, options);
            Regex regexLookFor = new Regex(_lookAroundParams.LookForPattern, options);

            for (int i = 0; i < _pageToMine.Length; i++)
            {
                LookAroundInteratedEventArgs eArgs = new LookAroundInteratedEventArgs(i);
                OnLookAroundIterated(eArgs);

                if (regexLookAround.IsMatch(_pageToMine[i]))
                {
                    switch (_lookAroundParams.DirectionTolook)
                    {
                        #region Iterative Logic
                        case LookAroundParams.LookDirection.Down:
                            {
                                for (int i2 = 0; i2 <= _lookAroundParams.LookAroundRadius; i2++)
                                {
                                    if (i + i2 >= _pageToMine.Length) break;

                                    Match match = regexLookFor.Match(_pageToMine[i + i2]);

                                    LookAroundInteratedEventArgs eArgs2 = new LookAroundInteratedEventArgs(i);
                                    OnLookAroundIterated(eArgs2);

                                    if (match.Success)
                                    {
                                        line = i + i2;
                                        return match;
                                    }
                                }

                                break;
                            }
                        case LookAroundParams.LookDirection.Up:
                            {
                                for (int i2 = 0; i2 <= _lookAroundParams.LookAroundRadius; i2++)
                                {
                                    if (i - i2 < 0) break; //Break if outside index bounds

                                    Match match = regexLookFor.Match(_pageToMine[i - i2]);

                                    LookAroundInteratedEventArgs eArgs2 = new LookAroundInteratedEventArgs(i - i2);
                                    OnLookAroundIterated(eArgs2);

                                    if (match.Success)
                                    {
                                        line = i - i2;
                                        return match;
                                    }
                                }

                                break;
                            }

                        case LookAroundParams.LookDirection.Both:
                            {
                                // Look Down
                                for (int i2 = 0; i2 <= _lookAroundParams.LookAroundRadius; i2++)
                                {
                                    if (i + i2 >= _pageToMine.Length) break;

                                    Match match = regexLookFor.Match(_pageToMine[i + i2]);

                                    LookAroundInteratedEventArgs eArgs2 = new LookAroundInteratedEventArgs(i + i2);
                                    OnLookAroundIterated(eArgs2);

                                    if (match.Success)
                                    {
                                        line = i + i2;
                                        return match;
                                    }
                                }

                                // look Up but this time start at i for efficiency, no need to
                                // scan the bottom half again
                                for (int i2 = i; i2 >= i - _lookAroundParams.LookAroundRadius; i2--)
                                {
                                    if (i2 == 0)
                                        break; //Break if outside index bounds

                                    Match match = regexLookFor.Match(_pageToMine[i2]);

                                    LookAroundInteratedEventArgs eArgs2 = new LookAroundInteratedEventArgs(i2);
                                    OnLookAroundIterated(eArgs2);

                                    if (match.Success)
                                    {
                                        line = i2;
                                        return match;
                                    }
                                }



                                break;
                            }

                        #endregion
                    }
                }
            }

            line = -1;
            return null;
        }
        #endregion

        public string OriginalPageLink
        {
            get { return _originalPageLink; }
        }

        public LookAroundParams LookAroundParameters
        {
            get { return _lookAroundParams; }
            set { _lookAroundParams = value; }
        }

        public string[] PageToMineContent
        {
            get { return _pageToMine; }
        }
    }
}
